(function() {
  var RequestBuilder, Router, collectPlugin, crema, disposable, plugins, pullPlugin, pushPlugin, _;

  crema = require("crema");

  RequestBuilder = require("./request").Builder;

  pushPlugin = require("./push/plugin");

  pullPlugin = require("./pull/plugin");

  collectPlugin = require("./collect/plugin");

  plugins = require("./plugins");

  disposable = require("disposable");

  _ = require("underscore");

  Router = (function() {
    /*
    */
    function Router() {
      this.directors = {};
      this.parse = crema;
      this._requestBuilder = new RequestBuilder(this);
      this._plugins = new plugins(this);
      this.use(pushPlugin);
      this.use(pullPlugin);
      this.use(collectPlugin);
    }

    /*
    	 uses a dispatcher
    */

    Router.prototype.use = function(plugin) {
      return this._plugins.add(plugin);
    };

    /*
    */

    Router.prototype.using = function() {
      return this._plugins.using();
    };

    /*
    	 listens for a request
    */

    Router.prototype.on = function(routeOrListeners, ops, callback) {
      var listenerDisposables, route, routes, type, _fn, _i, _len,
        _this = this;
      if (typeof ops === 'function') {
        callback = ops;
        ops = {};
      }
      listenerDisposables = disposable.create();
      if (typeof routeOrListeners === "object" && !callback) {
        for (type in routeOrListeners) {
          listenerDisposables.add(this.on(type, routeOrListeners[type]));
        }
        return listenerDisposables;
      }
      if (typeof routeOrListeners === "string") {
        routes = crema(routeOrListeners);
      } else if (routeOrListeners instanceof Array) {
        routes = routeOrListeners;
      } else {
        routes = [routeOrListeners];
      }
      _fn = function(route) {
        if (ops.type) route.type = ops.type;
        if (ops.tags) _.extend(route.tags, ops.tags);
        listenerDisposables.add(_this.director(route.type).addListener(route, callback));
        return _this._plugins.newListener({
          route: route,
          callback: callback
        });
      };
      for (_i = 0, _len = routes.length; _i < _len; _i++) {
        route = routes[_i];
        _fn(route);
      }
      return listenerDisposables;
    };

    /*
    	 returns the given director, or throws an error if it doesn't exist
    */

    Router.prototype.director = function(type) {
      var director;
      director = this.directors[type];
      if (!director) throw new Error("director " + type + " does not exist");
      return director;
    };

    /*
    */

    Router.prototype.paths = function(ops) {
      var director, name, paths;
      paths = [];
      for (name in this.directors) {
        director = this.directors[name];
        paths = paths.concat(director.paths(ops));
      }
      return paths;
    };

    Router.prototype.dispatch = function(requestWriter) {
      return this.director(requestWriter.type).dispatch(requestWriter);
    };

    /*
    	 abreviated
    */

    Router.prototype.req = function() {
      return this.request.apply(this, arguments);
    };

    /*
    	 Initializes a new request
    */

    Router.prototype.request = function(path, query, headers) {
      return this._requestBuilder.clean().path(typeof path === "string" ? crema.parsePath(path) : path).query(query).headers(headers);
    };

    return Router;

  })();

  module.exports = Router;

}).call(this);
